package org.unidal.eunit.model.transform;

import static org.unidal.eunit.model.Constants.ATTR_BEFORE_AFTER;
import static org.unidal.eunit.model.Constants.ATTR_ID;
import static org.unidal.eunit.model.Constants.ATTR_IGNORED;
import static org.unidal.eunit.model.Constants.ATTR_MESSAGE;
import static org.unidal.eunit.model.Constants.ATTR_NAME;
import static org.unidal.eunit.model.Constants.ATTR_PATTERN;
import static org.unidal.eunit.model.Constants.ATTR_RETURN_TYPE;
import static org.unidal.eunit.model.Constants.ATTR_STATIC;
import static org.unidal.eunit.model.Constants.ATTR_TEST;
import static org.unidal.eunit.model.Constants.ATTR_TIMEOUT;
import static org.unidal.eunit.model.Constants.ATTR_TYPE;
import static org.unidal.eunit.model.Constants.ELEMENT_GROUP;
import static org.unidal.eunit.model.Constants.ELEMENT_GROUPS;
import static org.unidal.eunit.model.Constants.ENTITY_EUNIT_CLASS;
import static org.unidal.eunit.model.Constants.ENTITY_EUNIT_EXCEPTION;
import static org.unidal.eunit.model.Constants.ENTITY_EUNIT_FIELD;
import static org.unidal.eunit.model.Constants.ENTITY_EUNIT_METHOD;
import static org.unidal.eunit.model.Constants.ENTITY_EUNIT_PARAMETER;
import static org.unidal.eunit.model.Constants.ENTITY_FIELDS;
import static org.unidal.eunit.model.Constants.ENTITY_METHODS;
import static org.unidal.eunit.model.Constants.ENTITY_PARAMETERS;

import java.lang.reflect.Array;
import java.util.Collection;

import org.unidal.eunit.model.IEntity;
import org.unidal.eunit.model.IVisitor;
import org.unidal.eunit.model.entity.EunitClass;
import org.unidal.eunit.model.entity.EunitException;
import org.unidal.eunit.model.entity.EunitField;
import org.unidal.eunit.model.entity.EunitMethod;
import org.unidal.eunit.model.entity.EunitParameter;

public class DefaultXmlBuilder implements IVisitor {

   private IVisitor m_visitor = this;

   private int m_level;

   private StringBuilder m_sb;

   private boolean m_compact;

   public DefaultXmlBuilder() {
      this(false);
   }

   public DefaultXmlBuilder(boolean compact) {
      this(compact, new StringBuilder(4096));
   }

   public DefaultXmlBuilder(boolean compact, StringBuilder sb) {
      m_compact = compact;
      m_sb = sb;
      m_sb.append("<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n");
   }

   public String buildXml(IEntity<?> entity) {
      entity.accept(m_visitor);
      return m_sb.toString();
   }

   protected void endTag(String name) {
      m_level--;

      indent();
      m_sb.append("</").append(name).append(">\r\n");
   }

   protected String escape(Object value) {
      return escape(value, false);
   }
   
   protected String escape(Object value, boolean text) {
      if (value == null) {
         return null;
      }

      String str = toString(value);
      int len = str.length();
      StringBuilder sb = new StringBuilder(len + 16);

      for (int i = 0; i < len; i++) {
         final char ch = str.charAt(i);

         switch (ch) {
         case '<':
            sb.append("&lt;");
            break;
         case '>':
            sb.append("&gt;");
            break;
         case '&':
            sb.append("&amp;");
            break;
         case '"':
            if (!text) {
               sb.append("&quot;");
               break;
            }
         default:
            sb.append(ch);
            break;
         }
      }

      return sb.toString();
   }
   
   protected void indent() {
      if (!m_compact) {
         for (int i = m_level - 1; i >= 0; i--) {
            m_sb.append("   ");
         }
      }
   }

   protected void startTag(String name) {
      startTag(name, false, null);
   }
   
   protected void startTag(String name, boolean closed, java.util.Map<String, String> dynamicAttributes, Object... nameValues) {
      startTag(name, null, closed, dynamicAttributes, nameValues);
   }

   protected void startTag(String name, java.util.Map<String, String> dynamicAttributes, Object... nameValues) {
      startTag(name, null, false, dynamicAttributes, nameValues);
   }

   protected void startTag(String name, Object text, boolean closed, java.util.Map<String, String> dynamicAttributes, Object... nameValues) {
      indent();

      m_sb.append('<').append(name);

      int len = nameValues.length;

      for (int i = 0; i + 1 < len; i += 2) {
         Object attrName = nameValues[i];
         Object attrValue = nameValues[i + 1];

         if (attrValue != null) {
            m_sb.append(' ').append(attrName).append("=\"").append(escape(attrValue)).append('"');
         }
      }

      if (dynamicAttributes != null) {
         for (java.util.Map.Entry<String, String> e : dynamicAttributes.entrySet()) {
            m_sb.append(' ').append(e.getKey()).append("=\"").append(escape(e.getValue())).append('"');
         }
      }

      if (text != null && closed) {
         m_sb.append('>');
         m_sb.append(escape(text, true));
         m_sb.append("</").append(name).append(">\r\n");
      } else {
         if (closed) {
            m_sb.append('/');
         } else {
            m_level++;
         }
   
         m_sb.append(">\r\n");
      }
   }

   @SuppressWarnings("unchecked")
   protected String toString(Object value) {
      if (value instanceof String) {
         return (String) value;
      } else if (value instanceof Collection) {
         Collection<Object> list = (Collection<Object>) value;
         StringBuilder sb = new StringBuilder(32);
         boolean first = true;

         for (Object item : list) {
            if (first) {
               first = false;
            } else {
               sb.append(',');
            }

            if (item != null) {
               sb.append(item);
            }
         }

         return sb.toString();
      } else if (value.getClass().isArray()) {
         int len = Array.getLength(value);
         StringBuilder sb = new StringBuilder(32);
         boolean first = true;

         for (int i = 0; i < len; i++) {
            Object item = Array.get(value, i);

            if (first) {
               first = false;
            } else {
               sb.append(',');
            }

            if (item != null) {
               sb.append(item);
            }
         }
		
         return sb.toString();
      }
 
      return String.valueOf(value);
   }

   protected void tagWithText(String name, String text, Object... nameValues) {
      if (text == null) {
         return;
      }
      
      indent();

      m_sb.append('<').append(name);

      int len = nameValues.length;

      for (int i = 0; i + 1 < len; i += 2) {
         Object attrName = nameValues[i];
         Object attrValue = nameValues[i + 1];

         if (attrValue != null) {
            m_sb.append(' ').append(attrName).append("=\"").append(escape(attrValue)).append('"');
         }
      }

      m_sb.append(">");
      m_sb.append(escape(text, true));
      m_sb.append("</").append(name).append(">\r\n");
   }

   protected void element(String name, String text, String defaultValue, boolean escape) {
      if (text == null || text.equals(defaultValue)) {
         return;
      }
      
      indent();
      
      m_sb.append('<').append(name).append(">");
      
      if (escape) {
         m_sb.append(escape(text, true));
      } else {
         m_sb.append("<![CDATA[").append(text).append("]]>");
      }
      
      m_sb.append("</").append(name).append(">\r\n");
   }

   @Override
   public void visitEunitClass(EunitClass eunitClass) {
      startTag(ENTITY_EUNIT_CLASS, null, ATTR_TYPE, eunitClass.getType() == null ? null : eunitClass.getType().getName(), ATTR_IGNORED, eunitClass.getIgnored());

      if (!eunitClass.getGroups().isEmpty()) {
         startTag(ELEMENT_GROUPS);

         for (String group : eunitClass.getGroups()) {
            tagWithText(ELEMENT_GROUP, group);
         }

         endTag(ELEMENT_GROUPS);
      }

      if (!eunitClass.getFields().isEmpty()) {
         startTag(ENTITY_FIELDS);

         for (EunitField field : eunitClass.getFields()) {
            field.accept(m_visitor);
         }

         endTag(ENTITY_FIELDS);
      }

      if (!eunitClass.getMethods().isEmpty()) {
         startTag(ENTITY_METHODS);

         for (EunitMethod method : eunitClass.getMethods()) {
            method.accept(m_visitor);
         }

         endTag(ENTITY_METHODS);
      }

      endTag(ENTITY_EUNIT_CLASS);
   }

   @Override
   public void visitEunitException(EunitException eunitException) {
      startTag(ENTITY_EUNIT_EXCEPTION, true, null, ATTR_TYPE, eunitException.getType() == null ? null : eunitException.getType().getName(), ATTR_MESSAGE, eunitException.getMessage(), ATTR_PATTERN, eunitException.getPattern());
   }

   @Override
   public void visitEunitField(EunitField eunitField) {
      startTag(ENTITY_EUNIT_FIELD, true, null, ATTR_NAME, eunitField.getName(), ATTR_TYPE, eunitField.getType() == null ? null : eunitField.getType().getName());
   }

   @Override
   public void visitEunitMethod(EunitMethod eunitMethod) {
      startTag(ENTITY_EUNIT_METHOD, null, ATTR_NAME, eunitMethod.getName(), ATTR_TEST, eunitMethod.getTest(), ATTR_TIMEOUT, eunitMethod.getTimeout(), ATTR_IGNORED, eunitMethod.getIgnored(), ATTR_BEFORE_AFTER, eunitMethod.getBeforeAfter(), ATTR_STATIC, eunitMethod.getStatic(), ATTR_RETURN_TYPE, eunitMethod.getReturnType() == null ? null : eunitMethod.getReturnType().getName());

      if (!eunitMethod.getGroups().isEmpty()) {
         startTag(ELEMENT_GROUPS);

         for (String group : eunitMethod.getGroups()) {
            tagWithText(ELEMENT_GROUP, group);
         }

         endTag(ELEMENT_GROUPS);
      }

      if (!eunitMethod.getParameters().isEmpty()) {
         startTag(ENTITY_PARAMETERS);

         for (EunitParameter parameter : eunitMethod.getParameters()) {
            parameter.accept(m_visitor);
         }

         endTag(ENTITY_PARAMETERS);
      }

      if (!eunitMethod.getExpectedExceptions().isEmpty()) {
         for (EunitException expectedException : eunitMethod.getExpectedExceptions()) {
            expectedException.accept(m_visitor);
         }
      }

      endTag(ENTITY_EUNIT_METHOD);
   }

   @Override
   public void visitEunitParameter(EunitParameter eunitParameter) {
      startTag(ENTITY_EUNIT_PARAMETER, true, null, ATTR_TYPE, eunitParameter.getType() == null ? null : eunitParameter.getType().getName(), ATTR_ID, eunitParameter.getId());
   }
}
